/*
 *  Pulse Induction Flip Coil Detector V01 by Joris Bijkerk
 *  
 *  Metal detector using Pulse Induction. 
 *  Instead of using two differend voltage levels for the pulse and the decay curve, the circuit uses two FETs to isolate the coil. 
 *  An addional resistor flips the polarity of the coil over in order to have the same polarity for the pulse and the decay curve
 *  
 *  V01
 *  Program is adapted to an 16x2 lcd display using a I2C converter.
 *  The calibration routine is called during start up. So if an recalibration is needed pressing the reset button would do the job.
 *  
 *  The Diplay indicates 3 values
 *  - Bargraph to indicate the strength of the signal
 *  - value indicating the difference between the calibration value and the current value. If this value is negative -> recalibrate
 *  - the current signal value -> usefull for setting up the detector (if unstable, retune the trim resistor.
 *  
 *  The speaker can be used aswell (or even as without the display, once the detector is tuned)
 *  
 */


#include <Wire.h> //I2C library
#include <LiquidCrystal_I2C.h>

// define some usefull constants 

#define max_sequence 1                              // maximum amount of different sequences (i. e. playing with different pulse times)
#define sequence_interval 100                       // multiplier for the pulste time in µS for each sequence
#define max_samples 10                              // amount of samples used to create the mean Value -> needs to fit to "sequence"
#define delayTreshold 0                             // defines a standard threashold to compensate for misreadings due to instable signal from the coil
#define baseDelay 350                               // defines a standard delay for the pulse
#define LedOnThreshold 2                            // defines the threshold where the LED, speaker and LCD diplay indicates something 

// define some usefull variables

volatile unsigned long timeStamp[max_sequence+1];     // stores the time difference between the end of the pulse signal and the intereupt at the comparator
volatile unsigned long timeMean[max_sequence+1];      // stores the mean value generated over "max_samples" values
volatile unsigned long timer1_overflow_count=0;     // stores how often the timer had an counter overflow -> not used
volatile int counter=0;                             // used for the counting the samples up to "max_samples"
volatile int sequence=0;                            // counter for diffent pulse sequences; 0=50µS, 1=150µS, 2=250µS, 3=350µS for baseDelay=50; max_sequence=3; sequence_interval=100 

int LedOn=0;                                        // indicates if the LED should indicate something -> defined as int to realize a hysteresis
int i;                                              // universal counter
unsigned long timeMeanRef[max_sequence+1];            // stores the refenece value aquired during the calibration  
int SumDif;                                         // diffenece between the reference Value and the measured value


unsigned char clockSelectBits = _BV(CS10);          // no prescale, full xtal
uint8_t oldSREG;

unsigned int filter[max_samples+1][max_sequence+1];     // array storing the samples per sequence -> for averaging the measured values 

// define some Pins

int PulsePin = 10;                                  // normal pulse pin for breadboard schematic
int PulsePin1 = 2;                                  // for the PCB schematic
int ledPin = 13;                                    // use the internal LED
int beepPin = 17;                                   // = A3


LiquidCrystal_I2C lcd(0x27, 2, 1, 0, 4, 5, 6, 7, 3, POSITIVE);  // Set the LCD I2C address

//SPECIAL CHARACTERS to be used for the bar graph indicator
byte p3[8] = {
        B10000,
        B10000,
        B10000,
        B10000,
        B10000,
        B10000,
        B10000,
        B10000
};

byte p4[8] = {
        B11000,
        B11000,
        B11000,
        B11000,
        B11000,
        B11000,
        B11000,
        B11000
};

byte p5[8] = {
        B11100,
        B11100,
        B11100,
        B11100,
        B11100,
        B11100,
        B11100,
        B11100
};

byte p6[8] = {
        B11110,
        B11110,
        B11110,
        B11110,
        B11110,
        B11110,
        B11110,
        B11110
};

byte p7[8] = {
        B11111,
        B11111,
        B11111,
        B11111,
        B11111,
        B11111,
        B11111,
        B11111
};


void setup() {

  pinMode(6, INPUT);                                // + of the comparator - by setting them as INPUT, they are set to high impedance
  pinMode(7, INPUT);                                // - of the comparator - by setting them as INPUT, they are set to high impedance

  cli();                            //stop interrupts

  TCCR1A = 0;                       // set entire TCCR1A register to 0
  TCCR1B = 0;                       // same for TCCR1B -> normal mode
  TCNT1  = 0;                       // initialize counter value to 0;
  TCCR1B |= clockSelectBits;        // sets prescaler and starts the clock
  // enable overflow interupt
  // sbi(TIMSK1, TOIE1);            // -> not required
  TIMSK1 = _BV(TOIE1);              // sets the timer overflow interrupt enable bit 

  sei();                            //allow interrupts

  ACSR =                            // setting the Comparator-Register  
    (0 << ACD) |                    // Analog Comparator: Enabled
    (0 << ACBG) |                   // Analog Comparator Bandgap Select: AIN0 is applied to the positive input
    (0 << ACO) |                    // Analog Comparator Output: Off
    (1 << ACI) |                    // Analog Comparator Interrupt Flag: Clear Pending Interrupt
    (1 << ACIE) |                   // Analog Comparator Interrupt: Enabled
    (0 << ACIC) |                   // Analog Comparator Input Capture: Disabled
    (0<<ACIS1) | (0<<ACIS0)         //interrupt on output toggle
    //(0<<ACIS1) | (1<<ACIS0)       //reserved
    //(1<<ACIS1) | (0<<ACIS0)       //interrupt on falling output edge
    //(1<<ACIS1) | (1<<ACIS0)       //interrupt on rising output edge
    ;   
  
  pinMode(ledPin, OUTPUT);          
  pinMode(PulsePin, OUTPUT);
  pinMode(PulsePin1, OUTPUT);
  pinMode(beepPin, OUTPUT);         
  pinMode(8, OUTPUT);               // used for the PCB Version of the detector -> providing 5V to the trim resistor               
  digitalWrite(8, HIGH);            // setting the 5V for the voltage devider on the trimmer -> for the PCB schematic 
  pinMode(A2, INPUT);               // required for bredboard, is connected to D6 -> setting to input will set it to high impedance
  
  lcd.begin(16,2);                  // inititate the LCD Display

// loading the bar graph characters to the 16x2 LCD controller

  lcd.createChar(1, p3);            // single bar character
  lcd.createChar(2, p4);            // double bar character
  lcd.createChar(3, p5);            // triple bar character
  lcd.createChar(4, p6);            // almost full character
  lcd.createChar(5, p7);            // full character
  
  delay(100);                       // stabelizing the voltage
  
  calibrate();                      // calibrating the response time of the coil
}


void calibrate(void){

  lcd.setCursor(0,0); 
  lcd.print("calibrating...  ");                          // telling what we are doing
  lcd.setCursor(0,1);
   
  for(i=0; i<(max_samples*max_sequence*10);i++){          // sequence -> how many different pulse times do we have; how many samples to we take per sequence; multiply by 10 to get stable readings
    pulse(baseDelay+sequence*sequence_interval);          // pulse with "baseDelay in µS + addional "sequnce_intererval" per sequence -> Values are written automatically to filter[][]  
    mean();                                               // create mean values
  }
  
  lcd.setCursor(0,1); 
  lcd.print("done!");                                     // telling that we are done
  delay(500);                                             // leave some time for reading
                                                          // create a usefull mean value
  for(i=0;i<max_sequence;i++){
    timeMeanRef[i]=timeMean[i];                           // use the generated mean values as a reference
  }
}

/*
 * loop() calls pulse() to generate a pulse, than analyses the if someting was found, activates the LED and the speaker
 */

void loop() { 
  
  pulse(baseDelay+sequence*sequence_interval);                       
  mean();                                                 // create the value for timeMean[]

  SumDif=0;
  for(i=0; i<max_sequence; i++)
    SumDif=SumDif+(timeStamp[i]-timeMeanRef[i]);          // adding the differences of all sequences aka different pulse times
  
  if(SumDif > delayTreshold){                             // creating a hysteresis
    LedOn++;                                              // by counting LedOn short misreadings can be filtered out
    if(LedOn>(LedOnThreshold+1))                          // smal hysteresis
      LedOn=(LedOnThreshold+1);
  } 
  else{
    LedOn--;
    if(LedOn<0)
      LedOn=0;
  }  
  if(LedOn>LedOnThreshold){                               // current hysteresis defined "LedOnThreshold", can be changed -> this influences stability and response time
    digitalWrite(ledPin, HIGH);  
    digitalWrite(beepPin, HIGH);                          // speaker on
    delayMicroseconds(SumDif);                            // this is "dirty" -> leads to a change in pulse frequency of the detector. Only relevant for stromg redings -> thus OK
    digitalWrite(beepPin, LOW);                           // speaker off
    LedOn=(LedOnThreshold+1);                             // just making sure
  }
  else
    digitalWrite(ledPin, LOW);
    
  if(counter==0){                                         // refreshing LCD Diplay only from time to time -> saving "overhead" in the pulse time
    draw();                                               // refresh display
  }
}


void pulse(int delaySec){
  digitalWrite(PulsePin,HIGH);                            // Set the output pin to high
  digitalWrite(PulsePin1,HIGH);
  delayMicroseconds(delaySec);                            // Delay e.g. 250 microseconds before setting low on output pin for 18WD@200mm or 28WD@100mm 
  resetTimer();                                           // reset the timer for interrupt
  digitalWrite(PulsePin,LOW);                             // Set low on output pin
  digitalWrite(PulsePin1,LOW);
  delay (5);                                              // just wait -> the counting is done by the interrupt -> this value influences the pulse frequency
  filter[counter][sequence]=timeStamp[sequence];          // timeStamp was automatically generated by the interupt
  sequence++;
  if(sequence == max_sequence){                           // restart at sequence 0
    sequence=0; 
    counter++;                                            // increase global counter for filter[][] array
    if (counter == max_samples)                           // restart at the samples 0
      counter=0;
  }
}


//************************ LCD Display Function *******************************************************
void draw(void) {
  lcd.clear();
  if(LedOn>LedOnThreshold){ 
    lcd.setCursor(0,0);
    lcd.write(5);                                   // at least on full character should be shown
    for(i=0; i<SumDif/10; i++){                     // how many further full characters should be shown
      lcd.write(5);
      if(i>13)                                      // limit the total amount of characters just to fill in the first line
        i=SumDif/10;
    }
    if(SumDif%10)                                   // is there a remainder?
      lcd.write((SumDif%10)/2);                     // show partial filled characters
  }
  lcd.setCursor(0,1);                               // second line
  lcd.print(SumDif);                                // difference between reference value and current reading
  lcd.setCursor(8,1); 
  lcd.print(timeStamp[0]);                          // current signal reading from the first sequnce -> can be changed to other sequences if more sequences are used
}

/*
 *  timer stuff
 */

// resetting the timer

void resetTimer(void){
  oldSREG = SREG;
  cli();                            // Disable interrupts
  TCNT1 = 0;                        //initialize counter value to 0;
  SREG = oldSREG;                   // Restore status register
  TCCR1B |= clockSelectBits;        // sets prescaler and starts the clock
  timer1_overflow_count=0;          // resets overflow counter
}

// ISR(ANALOG_COMP_vect) is called on interupt

ISR(ANALOG_COMP_vect) {
  timeStamp[sequence]=clicks();     // looks at the counter of timer1, much more sensetive than the micros() function
}

// ISR(TIMER1_OVF_vect) is called on counter overflow of timer1

ISR(TIMER1_OVF_vect){
  timer1_overflow_count++;
}

// clicks() is used to read the value of timer1

unsigned long clicks(void){
unsigned long tmp;
  oldSREG= SREG;
  cli();              
  tmp=TCNT1;              
  SREG = oldSREG;
  return tmp;
}

// created mean values for all sequences

void mean(void){
int j,i;                                                         // just another usefull counter
  for(j=0;j<max_sequence;j++)
    timeMean[j]=0;
  
  for(j=0;j<max_sequence;j++){                                   // e.g. <1                            
    for(i=0;i<max_samples;i++){                                  // e.g. <10
      timeMean[j]=timeMean[j]+filter[i][j];                      // filter[max_samples][max_sequence];
    }
    timeMean[j]=timeMean[j]/max_samples;   
  }
}
